#version 460 core

#define SHADOW
#define OCEAN_TINT

const vec3 RGB_DAMPING = vec3(1, 2, 4);

out vec4 o_colour;

in vec2 tex_coords; // (0, 0) is btm-left

uniform sampler2D screenTex;
uniform sampler2D depthTex;
uniform sampler2D shadowDepthTex;

uniform float b;
uniform vec3 cam_pos;
uniform vec3 light_pos;
uniform vec4 near_far;
uniform mat4 ip;
uniform mat4 iv;
uniform vec3 fog_colour;
uniform float max_fog_depth;
uniform float blur_max_depth;
uniform int blur_size = 1;
uniform float ct;
const float blur_distance = 50;
const float offset = 1.0 / 1000.0;

// volumetric params
uniform int volumetric_steps;
uniform mat4 lvp;
uniform float l_near;
uniform float l_far;
uniform float under_water_viewing_distance;
uniform vec3 water_tint_shallow;
uniform vec3 water_tint_deep;
uniform float water_tint_max_depth;
uniform float water_tint_exponent;
uniform float attenuation;
uniform float murk_exponent;

const vec3 water_shadow_colour = vec3(0.0, 0.05, 0.1);
const vec3 murk_colour = vec3(0.59, 0.74, 0.89);
const float water_shadow_intensity = 0.25;

float linear_depth(float depth) {
    float near = near_far.x;
    float far = near_far.y;
    float d = 2.0 * depth - 1.0;
    return 2.0 * near * far / (far + near - d * (far - near));
}

vec3 get_view_vector(vec2 pixel) {
    vec3 vv = vec3(ip * vec4(pixel * 2.0 - 1.0, 0.0, 1.0));
    return vec3(iv * vec4(vv, 0.0));
}

// From slides
vec4 world_pos_from_depth(float depth) {
    float z = depth * 2.0 - 1.0;
    vec4 clip_space_position = vec4(tex_coords * 2.0 - 1.0, z, 1.0);
    vec4 view_space_position = ip * clip_space_position;
    // Perspective divide
    view_space_position /= view_space_position.w;
    vec4 world_space_position = iv * view_space_position;
    return world_space_position;
}

// compute the height of the wave at that point in time (estimated, based on params in ocean.vert)
const vec4 wave1 = vec4(1, 1, 0.1, 41);
const vec4 wave2 = vec4(1, 0, 0.05, 21);
const vec4 wave3 = vec4(0, -1, 0.05, 17);
const float PI = 3.14159;

vec3 gwave_i(vec4 params, vec3 p) {
    float steepness = params.z;
    float wavelength = params.w;
    vec2 d = normalize(params.xy);
    float k = 2 * PI / wavelength;
    float c = sqrt(9.8 / k);
    float f = k * (dot(d, p.xz) - c * ct);
    float a = steepness / k;

    return vec3(d.x * a * cos(f), a * sin(f), d.y * a * cos(f));
}

vec3 gwave(vec3 ip) {
    vec3 p = ip;
    p.y = 0;
    p += gwave_i(wave1, ip);
    p += gwave_i(wave2, ip);
    p += gwave_i(wave3, ip);
    return p;
}

// given a point p, find the height of the ocean at that pt
float ocean_height(vec3 p) {
    // find where this point ends up
    vec3 tp = p;
    vec3 np = gwave(tp);
    // step backwards by displacement
    vec2 dis = np.xz - tp.xz;
    tp.xz = p.xz - dis;
    np = gwave(tp);
    // step backwards
    dis = np.xz - tp.xz;
    tp.xz = p.xz - dis;
    np = gwave(tp);

    return np.y;
}

// https://stackoverflow.com/questions/5149544/can-i-generate-a-random-number-inside-a-pixel-shader
float random(vec2 p)
{
    vec2 K1 = vec2(
        23.14069263277926, // e^pi (Gelfond's constant)
         2.665144142690225 // 2^sqrt(2) (Gelfond–Schneider constant)
    );
    return fract(cos(dot(p, K1)) * 12345.6789);
}

void main() {
    vec3 colour = texture(screenTex, tex_coords).rgb;
    float depth = texture(depthTex, tex_coords).r;

    vec3 view_vector = get_view_vector(tex_coords);
    vec3 cam_dir = normalize(view_vector);
    float scene_depth = linear_depth(depth) * length(view_vector);

    // box blur
    vec3 blur_colour = vec3(0.0);
    float num_samples = pow(blur_size * 2 + 1, 2);
    for (int i = -blur_size; i <= blur_size; i++) {
        for (int j = -blur_size; j <= blur_size; j++) {
            blur_colour += texture(screenTex, tex_coords + vec2(i, j) * offset).rgb / num_samples;
        }
    }

    float blur_mix = smoothstep(blur_max_depth, blur_max_depth + blur_distance, linear_depth(depth));
    colour = mix(colour, blur_colour, blur_mix);

    // calc under or over water
    bool is_under_water = cam_pos.y < ocean_height(cam_pos);

    vec4 world_pos = world_pos_from_depth(depth);

    #ifdef SHADOW
    // calc in shadow
    vec4 frag_pos_light_space = lvp *  world_pos;
    frag_pos_light_space = (clamp(frag_pos_light_space, -frag_pos_light_space.w, frag_pos_light_space.w) / frag_pos_light_space.w) * 0.5f + 0.5f;
    
    float ldf = frag_pos_light_space.z;
    float ldf_sampled = texture(shadowDepthTex, frag_pos_light_space.xy).r;
    bool is_in_shadow = !(ldf < ldf_sampled + 0.0001f) && !(ldf_sampled > 0.9999f);

    if(is_in_shadow) {
        colour = mix(colour, water_shadow_colour, water_shadow_intensity);
    }
    #endif

    if (is_under_water) {

        // VOLUMETRICS UNDER WATER

        // world space
        vec4 start_pos = vec4(cam_pos, 1.0); // Swapped start and end points because of distance limiting
        vec4 end_pos = world_pos; 
        vec4 world_dir = normalize(end_pos - start_pos);
        float world_step_size = clamp(length(end_pos - start_pos) / volumetric_steps, 0, under_water_viewing_distance / volumetric_steps); // Limit distance
        vec4 world_ray_pos = start_pos;

        float tint_strength = pow(clamp(length(end_pos - start_pos) / under_water_viewing_distance, 0, 1), water_tint_exponent);
        float volumetric_steps_i = 1.0 / volumetric_steps;
        float shaded = 0.0;
        float depth_sum = 0.0;
        float light_contribution = 0;
        
        float test = 0;

        // ray march
        for (int i = 0; i < volumetric_steps; i += 1) 
        {

            vec4 light_space_pos = lvp * world_ray_pos;
            vec3 light_uvz = (clamp(light_space_pos.xyz, -light_space_pos.w, light_space_pos.w) / light_space_pos.w) * 0.5f + 0.5f;
            float ld = light_uvz.z;
            float ld_sampled = texture(shadowDepthTex, light_uvz.xy).r;
            bool is_lit = (ld < ld_sampled + 0.0001f) || (ld_sampled > 0.9999f);
            
            //float ldw_sampled = texture(waterDepthTex, light_uvz.xy).r;
            //float d = clamp(ld - ldw_sampled, 0, 1) * l_far;
            float d = length(world_ray_pos.xyz - light_pos);
            float d_rcp = 1.0 / d;
            
            // murkyness += int(is_lit) * pow(volumetric_steps_i * attenuation * d_rcp * d_rcp, murk_exponent);
            // murkyness += int(is_lit) * volumetric_steps_i;
            shaded += int(!is_lit) * volumetric_steps_i;

            depth_sum += max(ocean_height(world_ray_pos.xyz) - world_ray_pos.y, 0);
            
            world_ray_pos += world_dir * world_step_size;
        }
        
        #ifdef OCEAN_TINT
        vec3 tint_color = mix(water_tint_shallow, water_tint_deep, clamp(depth_sum * volumetric_steps_i / water_tint_max_depth, 0, 1));
        colour = mix(colour, tint_color, clamp(pow(1 - shaded, murk_exponent), 0, 1) * tint_strength);        
        colour = mix(colour, tint_color, 
            clamp(pow(1 - shaded, murk_exponent) + int(length(end_pos - start_pos) > under_water_viewing_distance), 0, 1) * tint_strength
        );
        #endif
        
        //colour = mix(colour, water_shadow_colour, 1 - clamp(murkyness * murkyness_factor, 0, 1));
        //colour = mix(colour, water_shadow_colour, pow(shaded, murk_exponent));
        
    } else {
        // ABOVE WATER
        
        // simple atmosphere colouring
        vec3 lambda = exp(-b * min(max_fog_depth, scene_depth) * RGB_DAMPING);
        colour = mix(fog_colour, colour, lambda);

        // basic distance fog to hide edges
        //float fog_amt = 1.0 - exp(-scene_depth * b);
        //vec3 fog_colour = vec3(0.5, 0.6, 0.7);
        //colour = mix(colour, fog_colour, fog_amt);
    }

    o_colour = vec4(colour, 1.0);
}