#version 460 core

in vec3 position;
in vec3 normal;
in float ocean_ht;

out vec4 o_colour;

uniform sampler2D normal_tex;
uniform sampler2D caustics_tex;
uniform vec3 colour;
uniform vec3 sand_colour, dirt_colour;
uniform vec2 land_blend;
uniform float normal_blend;
uniform float land_scale;
uniform float ct;

uniform vec4 caustics_speed;
uniform vec4 caustics_scale1;
uniform vec4 caustics_scale2;
uniform float caustics_split;

struct Light {
    vec3 position;
    vec3 ambient;
    vec3 diffuse;
    vec3 specular;
};
uniform Light light;

vec3 get_normal_from_texture(vec2 uv) {
    vec3 rgb = texture(normal_tex, uv).rgb;
    return normalize(rgb * 2.0 - 1.0);
}

float saturate(float x) {
    return min(1.0, max(0.0, x));
}

vec4 hash4( vec2 p ) { return fract(sin(vec4( 1.0+dot(p,vec2(37.0,17.0)), 
                                              2.0+dot(p,vec2(11.0,47.0)),
                                              3.0+dot(p,vec2(41.0,29.0)),
                                              4.0+dot(p,vec2(23.0,31.0))))*103.0); 
}

vec3 textureNoTile(vec2 uv) {
    ivec2 iuv = ivec2(floor(uv));
    vec2 fuv = fract(uv);

    // generate per-tile transform
    vec4 ofa = hash4(iuv + ivec2(0, 0));
    vec4 ofb = hash4(iuv + ivec2(1, 0));
    vec4 ofc = hash4(iuv + ivec2(0, 1));
    vec4 ofd = hash4(iuv + ivec2(1, 1));
    
    // transform per-tile uvs
    ofa.zw = sign(ofa.zw - 0.5);
    ofb.zw = sign(ofb.zw - 0.5);
    ofc.zw = sign(ofc.zw - 0.5);
    ofd.zw = sign(ofd.zw - 0.5);
    
    // uvs
    vec2 uva = uv * ofa.zw + ofa.xy;
    vec2 uvb = uv * ofb.zw + ofb.xy;
    vec2 uvc = uv * ofc.zw + ofc.xy;
    vec2 uvd = uv * ofd.zw + ofd.xy;
        
    // fetch and blend
    vec2 b = smoothstep(0.25, 0.75, fuv);
    
    return mix( mix( get_normal_from_texture(uva), 
                     get_normal_from_texture(uvb), b.x), 
                mix( get_normal_from_texture(uvc),
                     get_normal_from_texture(uvd), b.x), b.y);
}


vec3 rnm(vec3 a, vec3 b) {
    a += vec3(0, 0, 1);
    b *= vec3(-1, -1, 1);
    return a * dot(a, b) / a.z - b;
}

void main() {
    vec3 light_dir = normalize(light.position - position);

    // compute triplanar mapping of normal map
    vec3 blend = abs(normal);
    blend = max(blend - 0.2, 0.0);
    blend /= dot(blend, vec3(1.0));

    vec2 uvx = position.zy / land_scale;
    vec2 uvy = position.xz / land_scale;
    vec2 uvz = position.xy / land_scale;

    vec3 tnx = textureNoTile(uvx);
    vec3 tny = textureNoTile(uvy);
    vec3 tnz = textureNoTile(uvz);

    vec3 avn = abs(normal);
    tnx = rnm(vec3(normal.zy, avn.x), tnx);
    tny = rnm(vec3(normal.xz, avn.y), tny);
    tnz = rnm(vec3(normal.xy, avn.z), tnz);

    vec3 asign = sign(normal);
    tnx.z *= asign.x;
    tny.z *= asign.y;
    tnz.z *= asign.z;

    vec3 tpnorm = normalize(tnx * blend.x + tny * blend.y + tnz * blend.z + normal);
    // vec3 tpnorm = normalize(textureNoTile(position.xz / land_scale));
    vec3 nn = normalize(mix(normal, tpnorm, normal_blend));

    vec3 ambient = light.ambient;
    float diff = max(dot(light_dir, nn), 0.0);
    vec3 diffuse = light.diffuse * diff;

    float sand_blend_ht = land_blend.x * (1 - land_blend.y);
    float sand_wt = saturate((position.y - sand_blend_ht) / (land_blend.x - sand_blend_ht));
    vec3 land_colour = mix(sand_colour, dirt_colour, sand_wt);

    vec3 result = (ambient + diffuse) * land_colour;

    // fake caustics using a texture
    if (position.y < ocean_ht) {
        vec2 uv1 = position.xz * caustics_scale1.xy + caustics_scale1.zw;
        uv1 += caustics_speed.xy * ct;
        vec2 uv2 = position.xz * caustics_scale2.xy + caustics_scale2.zw;
        uv2 += caustics_speed.zw * ct;

        float c1r = texture(caustics_tex, uv1 + vec2(caustics_split, caustics_split)).r;
        float c1g = texture(caustics_tex, uv1 + vec2(caustics_split, -caustics_split)).g;
        float c1b = texture(caustics_tex, uv1 + vec2(-caustics_split, -caustics_split)).b;

        float c2r = texture(caustics_tex, uv2 + vec2(caustics_split, caustics_split)).r;
        float c2g = texture(caustics_tex, uv2 + vec2(caustics_split, -caustics_split)).g;
        float c2b = texture(caustics_tex, uv2 + vec2(-caustics_split, -caustics_split)).b;

        vec3 caustics = min(vec3(c1r, c1g, c1b), vec3(c2r, c2g, c2b));
        result += caustics;
    }

    o_colour = vec4(result, 1.0);
}