#version 460 core

in vec3 position;
in vec3 normal;
in vec4 clip_space;
in vec2 tex_coords;

out vec4 o_colour;

uniform sampler2D reflectTex;
uniform sampler2D refractTex;
uniform sampler2D depthTex;
uniform sampler2D dudvTex;
uniform sampler2D normalTex;

uniform float normal_scaling;
uniform float normal_strength;
uniform vec3 colour;
uniform vec3 deep_colour;
uniform float deep_scale;
uniform vec3 cam_pos;
uniform vec4 near_far;
uniform float fresnel_strength;
uniform float ocean_mix;
uniform float dudv_strength;
uniform float wavespeed;
uniform float ct;
uniform float foam_height;
uniform vec3 foam_colour;

struct Light {
    vec3 position;
    vec3 ambient;
    vec3 diffuse;
    vec3 specular;
};
uniform Light light;

vec3 get_normal_from_texture(vec2 uv) {
    vec3 rgb = texture(normalTex, uv).rgb;
    return normalize(rgb * 2.0 - 1.0);
}

vec4 hash4( vec2 p ) { return fract(sin(vec4( 1.0+dot(p,vec2(37.0,17.0)), 
                                              2.0+dot(p,vec2(11.0,47.0)),
                                              3.0+dot(p,vec2(41.0,29.0)),
                                              4.0+dot(p,vec2(23.0,31.0))))*103.0); 
}

vec3 textureNoTile(vec2 uv) {
    ivec2 iuv = ivec2(floor(uv));
    vec2 fuv = fract(uv);

    // generate per-tile transform
    vec4 ofa = hash4(iuv + ivec2(0, 0));
    vec4 ofb = hash4(iuv + ivec2(1, 0));
    vec4 ofc = hash4(iuv + ivec2(0, 1));
    vec4 ofd = hash4(iuv + ivec2(1, 1));
    
    // transform per-tile uvs
    ofa.zw = sign(ofa.zw - 0.5);
    ofb.zw = sign(ofb.zw - 0.5);
    ofc.zw = sign(ofc.zw - 0.5);
    ofd.zw = sign(ofd.zw - 0.5);
    
    // uvs
    vec2 uva = uv * ofa.zw + ofa.xy;
    vec2 uvb = uv * ofb.zw + ofb.xy;
    vec2 uvc = uv * ofc.zw + ofc.xy;
    vec2 uvd = uv * ofd.zw + ofd.xy;
        
    // fetch and blend
    vec2 b = smoothstep(0.25, 0.75, fuv);
    
    return mix( mix( get_normal_from_texture(uva), 
                     get_normal_from_texture(uvb), b.x), 
                mix( get_normal_from_texture(uvc),
                     get_normal_from_texture(uvd), b.x), b.y);
}

float linear_depth(float depth) {
    float near = near_far.x;
    float far = near_far.y;
    float d = 2.0 * depth - 1.0;
    return 2.0 * near * far / (far + near - d * (far - near));
}

void main() {
    vec2 ndc = (clip_space.xy / clip_space.w) * 0.5 + 0.5;

    float depth = texture(depthTex, ndc).x;
    float cam_to_obj_distance = linear_depth(depth);
    float water_depth = gl_FragCoord.z;
    float cam_to_water_distance = linear_depth(water_depth);
    float underwater_depth = 0;
    if (cam_pos.y > 0) {
        underwater_depth = cam_to_obj_distance - cam_to_water_distance;
    } else {
        underwater_depth = cam_to_water_distance;
    }
    float ratio_underwater_to_max = underwater_depth / foam_height;

    float move_offset = fract(ct * wavespeed);
    vec2 dist1 = texture(dudvTex, vec2(tex_coords.x + move_offset, tex_coords.y)).rg * 2.0 - 1.0;
    vec2 dist2 = texture(dudvTex, vec2(tex_coords.x, tex_coords.y + move_offset)).rg * 2.0 - 1.0;
    vec2 tdist = dist1 + dist2;

    tdist *= dudv_strength;
    vec2 reflect_coords = vec2(ndc.x, -ndc.y) + tdist;
    vec2 refract_coords = ndc + tdist;

    reflect_coords.x = clamp(reflect_coords.x, 0.001, 0.999);
    reflect_coords.y = clamp(reflect_coords.y, -0.999, -0.001);
    refract_coords = clamp(refract_coords, 0.001, 0.999);

    // vec3 view_dir = normalize(cam_pos - position);
    // float fresnel = abs(dot(view_dir, normal));
    // vec3 mixed = mix(reflect_colour, refract_colour, pow(fresnel, fresnel_strength));

    vec3 reflect_colour = texture(reflectTex, reflect_coords).rgb;
    vec3 refract_colour = texture(refractTex, refract_coords).rgb;
    vec3 mixed = mix(reflect_colour, refract_colour, fresnel_strength);
    vec3 res_colour = mix(colour, mixed, ocean_mix);

    res_colour = mix(res_colour, deep_colour, clamp(underwater_depth / deep_scale, 0, 1));

    vec2 scaled_coords = normal_scaling * (position.xz * 0.5 + 0.5);
    vec2 normal_coords = vec2(scaled_coords.x + ct * wavespeed, scaled_coords.y - 0.3 * ct * wavespeed);
    vec3 tex_normal = normalize(textureNoTile(normal_coords));
    tex_normal = normalize(mix(normal, tex_normal, normal_strength));

    // vec3 ambient = light.ambient;

    vec3 light_dir = normalize(light.position - position);
    float diff = max(dot(light_dir, tex_normal), 0.0);
    vec3 diffuse = light.diffuse * diff;

    vec3 view_dir = normalize(cam_pos - position);
    // blinn-phong specular calculation
    vec3 halfway_dir = normalize(light_dir + view_dir);
    // vec3 reflect_dir = reflect(-light_dir, tex_normal);
    float spec = pow(max(dot(tex_normal, halfway_dir), 0.0), 32.0);
    vec3 specular = light.specular * spec;

    res_colour = diffuse * res_colour;
    float spec_str = clamp(pow(ratio_underwater_to_max, 4), 0, 1);

    // 0 at the most shallow, 1 at the deep end
    if (underwater_depth > 0 && underwater_depth < foam_height) {
        res_colour = mix(res_colour, foam_colour, 1 - ratio_underwater_to_max);
    }
    res_colour += spec_str * specular * res_colour;

    o_colour = vec4(res_colour, 1.0);
}